import re
import copy
import comsdk.graph as gr
from comsdk.edge import *
import importlib as imp
from typing import NamedTuple
from enum import Enum, auto


# Ключевые слова, описывающие граф
class Params():
    __slots__=(
        'module',
        'entry_func',
        'predicate',
        'selector',
        'function',
        'morphism',
        'parallelism',
        'comment'
    )
    def __init__(self):
        for slot in self.__slots__:
            setattr(self, slot, None)
    
    def __str__(self):
        stri = ""
        for s in self.__slots__:
            stri += s+": {}, ".format(getattr(self, s))
        return stri

class Func():
    __slots__ = (
        'module',
        'func', 
        'name'
    )
    def __init__(self, module, name, func=None):
        self.module = module
        self.name = name
        self.func = lambda data: data if func is None else func
    
    def __str__(self):
        return "{}.{}()".format(self.module, self.name)

class GraphFactory():
    __slots__ = (
        'states', 
        'graph',
        'tocpp'
    )
    def __init__(self, tocpp=False):
        self.states = {}
        self.tocpp = tocpp
    
    def add_state(self, statename, selectorname=None):
        if selectorname is not None:
            sp = entities[selectorname]
            if self.tocpp:
                sel_f = Func(sp.module, sp.entry_func)
            else:
                sel_f = Func(sp.module, sp.entry_func, getattr(imp.import_module(sp.module), sp.entry_func))
        if statename not in self.states:
            newstate = gr.State(statename, selector=sel_f)
            self.states[statename] = newstate
    
    def _create_morphism(self, morphname=None):
        if morphname is None:
            return Func("", ""), Func("", "")
        pred_f, func_f  = Func("",""), Func("", "")
        morph = entities[morphname]
        for m in morph.__slots__:
            if getattr(morph,m) != None:
                if m!="predicate" and m!="function":
                    print("ERROR: Morphisms could not have any params exept predicate and function!\n{}".format(morphname))
                    exit(0)
                if m=="predicate":
                    if getattr(morph,m) not in entities:
                        print("\tERROR: Predicate {} is not defined!".format(getattr(morph, m)))
                        exit(0)
                    pred = entities[getattr(morph, m)]
                    if self.tocpp:
                        pred_f = Func(pred.module, pred.entry_func)
                    else:
                        pred_f = Func(pred.module, pred.entry_func, getattr(imp.import_module(pred.module), pred.entry_func))
                if m=="function":
                    if getattr(morph,m) not in entities:
                        print("\tERROR: Function: {} is not defined!".format(getattr(morph, m)))
                        exit(0)
                    fu = entities[getattr(morph, m)]
                    if self.tocpp:
                        func_f = Func(fu.module, fu.entry_func)
                    else:
                        func_f = Func(fu.module, fu.entry_func, getattr(imp.import_module(fu.module), fu.entry_func))
        return pred_f, func_f


    def add_connection(self, st1, st2, morphism=None):
        pred, entr = self._create_morphism(morphism)
        self.states[st1].connect_to(self.states[st2], edge=Edge(pred, entr))
        print("{} --{}-{}--> {}".format(st1, pred, entr, st2))

    def build(self):
        self.graph = gr.Graph(self.states["__BEGIN__"], self.states["__END__"])
        self.graph.init_graph()

fact = GraphFactory(tocpp=True)
entities = {}

def check_brackets(rawfile):
    br = 0
    qu = 0
    for char in rawfile:
        if char == "[":
            br+=1
        elif char == "{":
            br+=1
        elif char == "(":
            br+=1
        elif char == "]":
            br-=1
        elif char == "}":
            br-=1
        elif char == ")":
            br-=1
        elif char =="\"":
            qu+=1
    if br!=0 or qu%2!=0:
        print("Brackets or quotes do not match! Check your file")
        exit(-1)

def split_multiple(param):
    res = []
    first=True
    for s in param.__slots__:
        attr = getattr(param,s)
        if attr is not None and '|' in attr:
            vals = attr.split('|')
            for v in vals:
                par = copy.copy(param)
                setattr(par, s, v)
                res.append(par)
    return res

    

#Props is line "[proFp=smth, ...]"
def param_from_props(props):
    parm = Params()
    props = props.replace("]", '')
    if '(' in props:
        #replaces , in (smth,smth) to ;
        props =props[:props.find('(')]+props[props.find('(')+1:props.find(')')].replace(',','|')
    rs =props.split(r",") #.split(r", ")
    for r in rs:
        r=r.split(r"=", 1)
        if r[0] in parm.__slots__:
            setattr(parm, r[0], r[1])
        else:
            print("\tERROR:Unknown parameter: "+ r[0])
            exit(-1)
    return parm

def param_from_entln(raw):
    res = re.split(r"\[", raw, 1)
    return res[0], param_from_props(res[1])


def topology(raw):
    spl = re.split(r"\s*(=>|->|\[|\])\s*", raw)
    spl = list(filter(lambda x: x!="[" and x!="]" and x!="", spl))
    left = spl[0].split(",")
    right = spl[2].split(",")
    if len(spl)>3:
        param_from_props(spl[3])
    if (len(left)>1) and (len(right)>1):
        print("ERROR:Ambigious multiple connection in line:\n\t{}".format(raw))
        exit()
    # many to one conection
    elif len(left)>1:
        p = param_from_props(spl[3])
        morphs = split_multiple(p)
        if len(morphs)!=len(left):
            print("\tERROR:Count of edges do not match to count of states in many to one connection!\n\t\t{}".format(raw))
            exit()
        fact.add_state(right[0])
        for i, st in enumerate(left):
            fact.add_state(st)
            fact.add_connection(st, right[0], morphs[i].morphism)
    # one to many connection, here could be selector
    elif len(right)>1:
        p = param_from_props(spl[3])
        fact.add_state(left[0],selector=p.selector)
        morphs = split_multiple(p)
        if len(morphs)!=len(right):
            print("\tERROR:Count of edges do not match to count of states in one to many connection!\n\t\t{}".format(raw))
            exit()
        for i, st in enumerate(right):
            fact.add_state(st)
            fact.add_connection(left[0], st, morphs[i].morphism)
    # one to one connection
    else:
        fact.add_state(left[0])
        fact.add_state(right[0])
        if len(spl)==4:
            pr =param_from_props(spl[3])
            fact.add_connection(left[0], right[0], pr.morphism)
        elif len(spl)==3:
            fact.add_connection(left[0], right[0], None)
    # print(spl)


file = open("./test.adot", "r")
dot = file.read()
check_brackets(dot)
#commts = re.findall(r"\".*\"", dot)
#
dot = re.sub(r"[ \t\r]", "", dot) #deleting all spaces
dot = re.sub(r"((digraph\w+\n?)|}|{)", "", dot)
dot = re.sub(r"\/\/.*", "", dot)
dot = re.sub(r"^\n$", "", dot)
#print("Checking graph...")
#graphcheck = re.search(r"\A(digraph)\w+\n?{((\/\/)*.*\n*)+}\Z", dot)
#if graphcheck is None:
#    print("Incorrect graph, check syntax!")
#    exit()
#
#print("Graph id good, processing!")
 

#dot = re.sub(r"//*$", "", dot) 
dotlines = dot.splitlines()
dotlines = list(filter(None, dotlines))


# ent_re - regular expr for edges, states, functions properties
ent_re = re.compile(r"^\w+\[.*\]$")
# top_re - regular expr for topology properties, most time consuming one
top_re = re.compile(r"^(\w+,?)+(->|=>)(\w+,?)+(\[(\w+=(\(?\w+,?\)?)+,?)+\])?")
# (r"^\w[\w\s,]*(->|=>)\s*\w[\w\s,=\[\]()]*$")

for i, ln in enumerate(dotlines):
    if ent_re.match(ln):
        name, parm = param_from_entln(ln)
        entities[name] = parm
    elif top_re.match(ln):
        topology(ln)

# for e in entities:
    # print(e, entities[e])
# 
# print("states:\n")
# for st in fact.states:
    # print(st)
# 
exit()

